# -*- coding: utf-8 -*-
"""
Created on Mon May 07 14:20:11 2018

@author: m.meliani

Adapted to new SMT version in march 2020 by Nathalie Bartoli
"""

try:
    import matplotlib

    matplotlib.use("Agg")
    NO_MATPLOTLIB = False
except ImportError:
    NO_MATPLOTLIB = True

import unittest
from copy import deepcopy

import numpy as np

from smt.applications.mfkpls import MFKPLS
from smt.problems import Sphere, TensorProduct
from smt.sampling_methods import LHS, FullFactorial
from smt.utils.misc import compute_relative_error
from smt.utils.silence import Silence
from smt.utils.sm_test_case import SMTestCase

print_output = False


class TestMFKPLS(SMTestCase):
    def setUp(self):
        self.nt = 100
        self.ne = 100
        self.ndim = 3
        self.n_comp = 2

    def test_mfkpls(self):
        self.problems = ["exp", "tanh", "cos"]

        for fname in self.problems:
            prob = TensorProduct(ndim=self.ndim, func=fname)
            sampling = FullFactorial(xlimits=prob.xlimits, clip=True)

            np.random.seed(0)
            xt = sampling(self.nt)
            yt = prob(xt)
            for i in range(self.ndim):
                yt = np.concatenate((yt, prob(xt, kx=i)), axis=1)

            y_lf = 2 * prob(xt) + 2
            x_lf = deepcopy(xt)
            np.random.seed(1)
            xe = sampling(self.ne)
            ye = prob(xe)

            # Modif MM
            sm = MFKPLS()

            if sm.options.is_declared("xlimits"):
                sm.options["xlimits"] = prob.xlimits
            sm.options["print_global"] = False

            # to test some options
            sm.options["eval_noise"] = True
            sm.options["optim_var"] = True

            # modif MM
            sm.options["n_comp"] = self.n_comp
            sm.options["theta0"] = [1e-2] * self.n_comp

            sm.set_training_values(xt, yt[:, 0])
            sm.set_training_values(x_lf, y_lf[:, 0], name=0)
            sm.set_training_values(xt, yt[:, 0])
            sm.set_training_values(x_lf, y_lf[:, 0], name=0)

            with Silence():
                sm.train()

            t_error = compute_relative_error(sm)
            e_error = compute_relative_error(sm, xe, ye)

            self.assert_error(t_error, 0.0, 1.5)
            self.assert_error(e_error, 0.0, 1.5)

    def test_mfkpls_derivs(self):
        if self.ndim < 2:
            print("To try test_mfkpls_derivs the dimension must be greater than 1")

        prob = Sphere(ndim=self.ndim)
        sampling = LHS(xlimits=prob.xlimits)

        # Modif MM
        nt = 100
        np.random.seed(0)
        xt = sampling(nt)
        yt = prob(xt)
        dyt = {}
        for kx in range(prob.xlimits.shape[0]):
            dyt[kx] = prob(xt, kx=kx)

        y_lf = 2 * prob(xt) + 2
        x_lf = deepcopy(xt)

        np.random.seed(1)
        xe = sampling(self.ne)

        dye = {}
        for kx in range(prob.xlimits.shape[0]):
            dye[kx] = prob(xe, kx=kx)
        print("n_comp mfkpls_deriv", self.n_comp)
        # modif MM
        sm = MFKPLS()

        if sm.options.is_declared("xlimits"):
            sm.options["xlimits"] = prob.xlimits
        sm.options["print_global"] = False

        # modif MM
        sm.options["n_comp"] = self.n_comp
        sm.options["theta0"] = [1e-2] * self.n_comp
        sm.set_training_values(xt, yt)
        sm.set_training_values(x_lf, y_lf, name=0)

        with Silence():
            sm.train()

        e_error0 = compute_relative_error(sm, xe, dye[0], 0)
        e_error1 = compute_relative_error(sm, xe, dye[1], 1)

        self.assert_error(e_error0, 0.0, 1e-1)
        self.assert_error(e_error1, 0.0, 1e-1)

    @staticmethod
    def run_mfkpls_example():
        import matplotlib.pyplot as plt
        import numpy as np

        from smt.applications.mfk import NestedLHS
        from smt.applications.mfkpls import MFKPLS

        # low fidelity model
        def lf_function(x):
            import numpy as np

            return (
                0.5 * ((x * 6 - 2) ** 2) * np.sin((x * 6 - 2) * 2)
                + (x - 0.5) * 10.0
                - 5
            )

        # high fidelity model
        def hf_function(x):
            import numpy as np

            return ((x * 6 - 2) ** 2) * np.sin((x * 6 - 2) * 2)

        # Problem set up
        xlimits = np.array([[0.0, 1.0]])
        xdoes = NestedLHS(nlevel=2, xlimits=xlimits, random_state=0)
        xt_c, xt_e = xdoes(7)

        # Evaluate the HF and LF functions
        yt_e = hf_function(xt_e)
        yt_c = lf_function(xt_c)

        # choice of number of PLS components
        ncomp = 1
        sm = MFKPLS(n_comp=ncomp, theta0=ncomp * [1.0])

        # low-fidelity dataset names being integers from 0 to level-1
        sm.set_training_values(xt_c, yt_c, name=0)
        # high-fidelity dataset without name
        sm.set_training_values(xt_e, yt_e)

        # train the model
        sm.train()

        x = np.linspace(0, 1, 101, endpoint=True).reshape(-1, 1)

        # query the outputs
        y = sm.predict_values(x)
        _mse = sm.predict_variances(x)
        _derivs = sm.predict_derivatives(x, kx=0)

        plt.figure()

        plt.plot(x, hf_function(x), label="reference")
        plt.plot(x, y, linestyle="-.", label="mean_gp")
        plt.scatter(xt_e, yt_e, marker="o", color="k", label="HF doe")
        plt.scatter(xt_c, yt_c, marker="*", color="g", label="LF doe")

        plt.legend(loc=0)
        plt.ylim(-10, 17)
        plt.xlim(-0.1, 1.1)
        plt.xlabel(r"$x$")
        plt.ylabel(r"$y$")

        plt.show()

    # run scripts are used in documentation as documentation is not always rebuild
    # make a test run by pytest to test the run scripts
    @unittest.skipIf(NO_MATPLOTLIB, "Matplotlib not installed")
    def test_run_mfkpls_example(self):
        self.run_mfkpls_example()


if __name__ == "__main__":
    unittest.main()



"""
Author: Hugo Reimeringer <hugo.reimeringer@onera.fr>
"""

import unittest
from smt.utils.sm_test_case import SMTestCase
import numpy as np
from scipy import special
from smt.applications import PODI
from smt.sampling_methods import LHS


def cos_coeff(i: int, x: np.ndarray):
    """Generates the i-th coefficient for the one-dimension problem."""

    a = 2 * i % 2 - 1
    return a * x[:, 0] * np.cos(i * x[:, 0])


def cos_coeff_nd(i: int, x: np.ndarray):
    """Generates the i-th coefficient for the multi dimensions problem."""
    a = 2 * i % 2 - 1
    return a * sum(x.T) * np.cos(i * sum(x.T))


def Legendre(i: int, t: np.ndarray):
    """Generates the i-th Legendre's polynom and returns its values at input values."""

    return special.legendre(i)(t)


class Test(SMTestCase):
    """Class to test Proper Orthogonal Decomposition and Interpolation (PODI) surrogate models based."""

    def setUp(self):
        """Sets up the test case for the tests."""

        self.full_database = self.pb_1d()
        self.database = self.full_database[:, : self.nt]

    @staticmethod
    def gram_schmidt(input_array: np.ndarray) -> np.ndarray:
        """Static method that performs the  Gram-Schmidt's algorithm."""

        basis = np.zeros_like(input_array)
        for i in range(len(input_array)):
            basis[i] = input_array[i]
            for j in range(i):
                basis[i] -= (
                    np.dot(input_array[i], basis[j])
                    / np.dot(basis[j], basis[j])
                    * basis[j]
                )
            basis[i] /= np.linalg.norm(basis[i])
        return basis

    @staticmethod
    def check_projection(
        basis_original: np.ndarray, basis_pod: np.ndarray
    ) -> np.ndarray:
        """
        Computes the residue's norm of the projection of the pod basis on the subspace generated by the problem basis.

        Parameters
        ----------
        basis_original : np.ndarray

        basis_pod : np.ndarray

        Returns
        -------
        norm_residue : np.ndarray
            norm of the left residue
        """

        norm_residue = np.zeros(basis_pod.shape[1])

        projection = basis_original.dot(np.dot(basis_original.T, basis_pod))

        for i in range(projection.shape[1]):
            proj = projection[:, i]
            norm_residue[i] = np.linalg.norm(basis_pod[:, i] - proj)
        return norm_residue

    def pb_1d(self, nt=40, nv=20) -> np.ndarray:
        """
        Constructs the one-dimension problem

        Parameters
        ----------
        nt : int
            Number of training values desired (number of snapshot).

        Returns
        -------
        database : np.ndarray
            Snapshot matrix, each row corresponds to the values of our problem at a specific snapshot.
        """
        self.seed = 42

        self.ny = 100
        self.t = np.linspace(-1, 1, self.ny)
        self.n_modes_test = 10

        xlimits = np.array([[0, 4]])
        sampling = LHS(xlimits=xlimits, random_state=self.seed)
        self.nt = nt
        self.xt = sampling(self.nt)
        self.nn = 15
        self.xn = sampling(self.nn)
        self.nv = nv
        sampling_new = LHS(xlimits=xlimits, random_state=self.seed + 1)
        self.xv = sampling_new(self.nv)
        self.x = np.concatenate((self.xt, self.xv))

        u0 = np.zeros((self.ny, 1))

        alpha = np.zeros((self.x.shape[0], self.n_modes_test))
        for i in range(self.n_modes_test):
            alpha[:, i] = cos_coeff(i, self.x)

        V_init = np.zeros((self.ny, self.n_modes_test))
        for i in range(self.n_modes_test):
            V_init[:, i] = Legendre(i, self.t)

        V = Test.gram_schmidt(V_init.T).T
        database = u0 + np.dot(V, alpha.T)
        self.basis_original = V

        return database

    def pb_2d_local(self, nt1=10, nt2=10) -> np.ndarray:
        """
        Constructs the two-dimension problem adapted for local POD (can be used for global POD as well).

        Parameters
        ----------
        nt1 : int
            Number of values for the first dimension of the inputs.
            Each one of these values will correspond to a line in the DoE.
        nt2 : int
            Number of values for the second dimension of the inputs.

        Returns
        -------
        database : np.ndarray
            Snapshot matrix, each row corresponds to the values of our problem at a specific snapshot.
        """
        self.seed = 42

        self.ny = 100
        self.t = np.linspace(-1, 1, self.ny)
        self.n_modes_test = 10

        xlimits = [[0, 1], [0, 4]]
        sampling_x1 = LHS(xlimits=np.array([xlimits[0]]), random_state=self.seed)
        sampling_x2 = LHS(xlimits=np.array([xlimits[1]]), random_state=self.seed + 1)

        self.nt1 = nt1
        self.nt2 = nt2
        self.nt = self.nt1 * self.nt2
        self.xt1 = sampling_x1(self.nt1)
        self.xt2 = sampling_x2(self.nt)
        self.xt = np.zeros((self.nt, 2))
        self.xt[:, 1] = self.xt2[:, 0]
        for i, elt in enumerate(self.xt1):
            self.xt[i * self.nt2 : (i + 1) * self.nt2, 0] = elt

        sampling_new = LHS(xlimits=np.array(xlimits), random_state=self.seed)

        self.nn = 15
        self.xn = sampling_new(self.nn)
        self.nv = 10
        self.xv = sampling_new(self.nv)
        self.x = np.concatenate((self.xt, self.xv))

        u0 = np.zeros((self.ny, 1))

        alpha = np.zeros((self.x.shape[0], self.n_modes_test))
        for i in range(self.n_modes_test):
            alpha[:, i] = cos_coeff_nd(i, self.x)

        V_init = np.zeros((self.ny, self.n_modes_test))
        for i in range(self.n_modes_test):
            V_init[:, i] = Legendre(i, self.t)

        V = Test.gram_schmidt(V_init.T).T
        database = u0 + np.dot(V, alpha.T)
        self.basis_original = V

        return database

    def test_predict(self):
        """Tests the predict methods."""

        sm = PODI()

        sm.compute_pod(self.database, tol=1, seed=self.seed)
        sm.set_interp_options("KRG")
        sm.set_training_values(self.xt)

        for predict_method in [sm.predict_values, sm.predict_variances]:
            error_msg = f"It should not be possible to call {predict_method.__name__} before the training."
            with self.assertRaises(RuntimeError, msg=error_msg):
                predict_method(self.xn)

        for predict_method in [sm.predict_derivatives, sm.predict_variance_derivatives]:
            error_msg = f"It should not be possible to call {predict_method.__name__} before the training."
            with self.assertRaises(RuntimeError, msg=error_msg):
                predict_method(self.xn, 0)

        sm.train()

        error_msg = "It should not be possible to make a prediction with incorrect dimension input."
        for predict_method in [sm.predict_values, sm.predict_variances]:
            with self.assertRaises(ValueError, msg=error_msg):
                predict_method(np.ones((1, self.xn.shape[1] + 1)))

        for predict_method in [sm.predict_derivatives, sm.predict_variance_derivatives]:
            with self.assertRaises(ValueError, msg=error_msg):
                predict_method(np.ones((1, self.xn.shape[1] + 1)), 0)

        error_msg = "It should not be possible to predict a derivative out of the input's dimensions."
        for predict_method in [sm.predict_derivatives, sm.predict_variance_derivatives]:
            with self.assertRaises(ValueError, msg=error_msg):
                predict_method(self.xn, self.xn.shape[1] + 1)

        var_xt = sm.predict_variances(self.xt)

        np.testing.assert_allclose(var_xt, np.zeros(var_xt.shape), atol=1e-6)

        mean_xn = sm.predict_values(self.xn)
        var_xn = sm.predict_variances(self.xn)
        deriv_xn = sm.predict_derivatives(self.xn, 0)
        var_deriv_xn = sm.predict_variance_derivatives(self.xn, 0)

        self.assertEqual(mean_xn.shape, (self.ny, self.nn))
        self.assertEqual(var_xn.shape, (self.ny, self.nn))
        self.assertEqual(deriv_xn.shape, (self.ny, self.nn))
        self.assertEqual(var_deriv_xn.shape, (self.ny, self.nn))

        mean_xv = sm.predict_values(self.xv)

        diff = self.full_database[:, self.nt :] - mean_xv
        rms_error = [np.sqrt(np.mean(diff[:, i] ** 2)) for i in range(diff.shape[1])]

        np.testing.assert_allclose(rms_error, np.zeros(self.nv), atol=1e-2)

    def test_set_options(self):
        """Tests the method that sets the interpolations options settings."""

        sm = PODI()

        error_msg = "It should not be possible to call 'set_interp_options' before 'compute_pod'."
        with self.assertRaises(RuntimeError, msg=error_msg):
            sm.set_interp_options()

        sm.compute_pod(self.database, n_modes=2, seed=self.seed)

        error_msg = (
            "It should not be possible to initialize an unavailable surrogate model."
        )
        with self.assertRaises(ValueError, msg=error_msg):
            sm.set_interp_options("non existing surrogate model")

        error_msg = (
            "It should not be possible to use a non-valid size for the list of options."
        )
        with self.assertRaises(ValueError, msg=error_msg):
            sm.set_interp_options("KRG", [{}, {}, {}])

        options_global = [
            {
                "poly": "quadratic",
                "corr": "matern32",
                "pow_exp_power": 0.38,
                "theta0": [1e-1],
            }
        ]
        sm.set_interp_options("KRG", options_global)

        sm_list = sm.get_interp_coeff()
        for interp_coeff in sm_list:
            for key in options_global[0].keys():
                self.assertEqual(interp_coeff.options[key], options_global[0][key])

        options_local = [{"poly": "quadratic"}, {"corr": "matern52"}]
        sm.set_interp_options("KRG", options_local)

        sm_list = sm.get_interp_coeff()
        for i, interp_coeff in enumerate(sm_list):
            for key in options_local[i].keys():
                self.assertEqual(interp_coeff.options[key], options_local[i][key])

    def test_pod(self):
        """Tests the computing of the global pod."""

        sm = PODI()

        error_msg = "It should not be possible to execute compute_pod without tol or n_modes argument."
        with self.assertRaises(ValueError, msg=error_msg):
            sm.compute_pod(self.database, seed=self.seed)

        error_msg = "It should not be possible to execute compute_pod with both tol and n_modes arguments."
        with self.assertRaises(ValueError, msg=error_msg):
            sm.compute_pod(self.database, tol=0.1, n_modes=1, seed=self.seed)

        error_msg = "It should not be possible to execute compute_pod with more mods than data values."
        with self.assertRaises(ValueError, msg=error_msg):
            sm.compute_pod(self.database, n_modes=self.nt + 1, seed=self.seed)

        sm.compute_pod(self.database, tol=1, seed=self.seed)
        self.assertEqual(sm.get_ev_ratio(), 1)

        n_modes = sm.get_n_modes()
        self.assertLessEqual(n_modes, self.n_modes_test)

        basis_pod = sm.get_singular_vectors()

        self.assertEqual(basis_pod.shape, (self.ny, min(self.nt, self.ny)))

        singular_values = sm.get_singular_values()
        self.assertEqual(len(singular_values), min(self.nt, self.ny))
        np.testing.assert_allclose(
            singular_values[n_modes:],
            np.zeros(min(self.nt, self.ny) - n_modes),
            atol=1e-6,
        )

        norm_residue = Test.check_projection(self.basis_original, basis_pod)
        np.testing.assert_allclose(norm_residue[:n_modes], np.zeros(n_modes), atol=1e-6)

        # local pod
        error_msg = "It should not be possible to compute the local pod with a local basis with incorrect dimensions."
        with self.assertRaises(ValueError, msg=error_msg):
            sm.compute_pod(
                database=self.database, pod_type="local", local_basis=basis_pod[1:, :]
            )
        sm.compute_pod(database=self.database, pod_type="local", local_basis=basis_pod)

    def test_interp_subspaces(self):
        xt1 = np.array([[1], [2]])

        m1 = np.array([[1, 1], [1, 2]])
        m2 = np.array([[1, 1], [2, 1]])
        input_matrices = [m1, m2]

        xn1 = np.array([[1.5]])

        error_msg = "It should not be possible to interpolate subspaces if the bases don't have the same dimensions."
        with self.assertRaises(ValueError, msg=error_msg):
            PODI.interp_subspaces(xt1=xt1, input_matrices=[m1, m2[:1]], xn1=xn1)

        error_msg = "It should not be possible to interpolate if there are not as much bases than assiociated values."
        with self.assertRaises(ValueError, msg=error_msg):
            PODI.interp_subspaces(xt1=xt1[0], input_matrices=input_matrices, xn1=xn1)

        output_list = PODI.interp_subspaces(
            xt1=xt1, input_matrices=input_matrices, xn1=xn1, print_global=False
        )
        self.assertEqual(len(output_list), xn1.shape[0])
        for basis in output_list:
            self.assertEqual(basis.shape, input_matrices[0].shape)

        # with realizations
        n_realizations = 10
        output_list, realization_list = PODI.interp_subspaces(
            xt1=xt1,
            input_matrices=input_matrices,
            xn1=xn1,
            compute_realizations=True,
            n_realizations=n_realizations,
            print_global=False,
        )
        self.assertEqual(len(output_list), xn1.shape[0])
        for basis in output_list:
            self.assertEqual(basis.shape, input_matrices[0].shape)
        self.assertEqual(len(realization_list), xn1.shape[0])
        for realization in realization_list:
            self.assertEqual(len(realization), n_realizations)
            for basis in realization:
                self.assertEqual(basis.shape, input_matrices[0].shape)

    def test_set_training_train(self):
        """Tests the set_training_values and train methods."""
        sm = PODI()

        error_msg = (
            "It should not be possible to set training values before computing the pod."
        )
        with self.assertRaises(RuntimeError, msg=error_msg):
            sm.set_training_values(self.xt)

        error_msg = (
            "It should not be possible to train the model before computing the pod."
        )
        with self.assertRaises(RuntimeError, msg=error_msg):
            sm.train()

        sm.compute_pod(self.database, n_modes=1, seed=self.seed)

        error_msg = (
            "It should not be possible to set training values with incorrect size."
        )
        with self.assertRaises(ValueError, msg=error_msg):
            sm.set_training_values(self.xt[1:])

        error_msg = (
            "It should not be possible to train the model before setting train values."
        )
        with self.assertRaises(RuntimeError, msg=error_msg):
            sm.train()

    @staticmethod
    def run_podi_example_1d_global():
        import matplotlib.pyplot as plt
        import numpy as np
        from smt.applications import PODI
        from smt.sampling_methods import LHS

        light_pink = np.array((250, 233, 232)) / 255

        p = 100
        t = np.linspace(-1, 1, p)
        n_modes_test = 10

        def function_test_1d(x, t, n_modes_test, p):
            import numpy as np  # Note: only required by SMT doc testing toolchain

            def cos_coeff(i: int, x: np.ndarray):
                a = 2 * i % 2 - 1
                return a * x[:, 0] * np.cos(i * x[:, 0])

            def Legendre(i: int, t: np.ndarray):
                from scipy import special

                return special.legendre(i)(t)

            def gram_schmidt(input_array: np.ndarray) -> np.ndarray:
                """To perform the  Gram-Schmidt's algorithm."""

                basis = np.zeros_like(input_array)
                for i in range(len(input_array)):
                    basis[i] = input_array[i]
                    for j in range(i):
                        basis[i] -= (
                            np.dot(input_array[i], basis[j])
                            / np.dot(basis[j], basis[j])
                            * basis[j]
                        )
                    basis[i] /= np.linalg.norm(basis[i])
                return basis

            u0 = np.zeros((p, 1))

            alpha = np.zeros((x.shape[0], n_modes_test))
            for i in range(n_modes_test):
                alpha[:, i] = cos_coeff(i, x)

            V_init = np.zeros((p, n_modes_test))
            for i in range(n_modes_test):
                V_init[:, i] = Legendre(i, t)

            V = gram_schmidt(V_init.T).T
            database = u0 + np.dot(V, alpha.T)

            return database

        seed_sampling = 42
        xlimits = np.array([[0, 4]])
        sampling = LHS(xlimits=xlimits, random_state=seed_sampling)

        nt = 40
        xt = sampling(nt)

        nv = 50
        xv = sampling(nv)

        x = np.concatenate((xt, xv))
        dbfull = function_test_1d(x, t, n_modes_test, p)

        # Training data
        dbt = dbfull[:, :nt]

        # Validation data
        dbv = dbfull[:, nt:]

        podi = PODI()
        seed_pod = 42
        podi.compute_pod(dbt, tol=0.9999, seed=seed_pod)
        podi.set_training_values(xt)
        podi.train()

        values = podi.predict_values(xv)
        variances = podi.predict_variances(xv)

        # computing the POD errors:
        # [max_interp_error, max_proj_error, max_total_error] = PODI.compute_pod_errors(xt = xt, database = dbt)
        # print("interpolation error: ", max_interp_error)
        # print("projection error: ", max_proj_error)
        # print("total error: ", max_total_error)

        # Choosing a value from the validation inputs
        i = nv // 2

        diff = dbv[:, i] - values[:, i]
        rms_error = np.sqrt(np.mean(diff**2))
        plt.figure(figsize=(8, 5))
        light_pink = np.array((250, 233, 232)) / 255
        plt.fill_between(
            np.ravel(t),
            np.ravel(values[:, i] - 3 * np.sqrt(variances[:, i])),
            np.ravel(values[:, i] + 3 * np.sqrt(variances[:, i])),
            color=light_pink,
            label="confiance interval (99%)",
        )
        plt.scatter(
            t,
            values[:, i],
            color="r",
            marker="x",
            s=15,
            alpha=1.0,
            label="prediction (mean)",
        )
        plt.scatter(
            t,
            dbv[:, i],
            color="b",
            marker="*",
            s=5,
            alpha=1.0,
            label="reference",
        )
        plt.plot([], [], color="w", label="rms = " + str(round(rms_error, 9)))

        ax = plt.gca()
        ax.axes.xaxis.set_visible(False)

        plt.ylabel("u(x = " + str(xv[i, 0])[:4] + ")")
        plt.title("Estimation of u at x = " + str(xv[i, 0])[:4])
        plt.legend()
        plt.show()

    @staticmethod
    def run_podi_example_2d_local():
        import matplotlib.pyplot as plt
        import numpy as np
        from smt.applications import PODI
        from smt.sampling_methods import LHS

        p = 100
        y = np.linspace(-1, 1, p)
        n_modes_test = 10

        def function_test_2d_local(x, y, n_modes_test, p):
            import numpy as np  # Note: only required by SMT doc testing toolchain

            def cos_coeff_nd(i: int, x: np.ndarray):
                a = 2 * i % 2 - 1
                return a * sum(x.T) * np.cos(i * sum(x.T))

            def Legendre(i: int, y: np.ndarray):
                from scipy import special

                return special.legendre(i)(y)

            def gram_schmidt(input_array: np.ndarray) -> np.ndarray:
                """To perform the  Gram-Schmidt's algorithm."""

                basis = np.zeros_like(input_array)
                for i in range(len(input_array)):
                    basis[i] = input_array[i]
                    for j in range(i):
                        basis[i] -= (
                            np.dot(input_array[i], basis[j])
                            / np.dot(basis[j], basis[j])
                            * basis[j]
                        )
                    basis[i] /= np.linalg.norm(basis[i])
                return basis

            u0 = np.zeros((p, 1))

            alpha = np.zeros((x.shape[0], n_modes_test))
            for i in range(n_modes_test):
                alpha[:, i] = cos_coeff_nd(i, x)

            V_init = np.zeros((p, n_modes_test))
            for i in range(n_modes_test):
                V_init[:, i] = Legendre(i, y)

            V = gram_schmidt(V_init.T).T
            database = u0 + np.dot(V, alpha.T)

            return database

        seed = 42
        xlimits = [[0, 1], [0, 4]]
        sampling_x1 = LHS(xlimits=np.array([xlimits[0]]), random_state=seed)
        sampling_x2 = LHS(xlimits=np.array([xlimits[1]]), random_state=seed + 1)

        nt1 = 25
        nt2 = 10
        nt = nt1 * nt2
        xt1 = sampling_x1(nt1)
        xt2 = sampling_x2(nt)
        xt = np.zeros((nt, 2))
        xt[:, 1] = xt2[:, 0]
        for i, elt in enumerate(xt1):
            xt[i * nt2 : (i + 1) * nt2, 0] = elt

        sampling_new = LHS(xlimits=np.array(xlimits), random_state=seed)

        nv = 15
        xv = sampling_new(nv)
        xv1 = np.atleast_2d(xv[:, 0]).T
        x = np.concatenate((xt, xv))

        dbfull = function_test_2d_local(x, y, n_modes_test, p)

        # Training data
        dbt = dbfull[:, :nt]

        # Validation data
        dbv = dbfull[:, nt:]

        plt.figure(figsize=(8, 5))
        axes = plt.gca()
        axes.tick_params(axis="x", labelsize=14)
        axes.tick_params(axis="y", labelsize=14)
        plt.scatter(xt[:, 1], xt[:, 0], marker="x", label="Training points", color="g")
        plt.scatter(
            xv[:, 1], xv[:, 0], marker="*", label="Validation points", color="r"
        )
        plt.xlabel(r"$x^{(2)}$", fontsize=18)
        plt.ylabel(r"$x^{(1)}$", fontsize=18)
        plt.legend(loc="lower left", fontsize=14)

        tol = 0.9999  # SVD tolerance for each line's POD basis
        local_pod_bases = []  # list of each line's POD bases
        n_modes_list = []

        podi = PODI()
        for i in range(nt1):
            db_loc = dbt[:, i * nt2 : (i + 1) * nt2]
            podi.compute_pod(
                db_loc, pod_type="global", n_modes=min(db_loc.shape), seed=i
            )
            ev_list = podi.get_ev_list()

            n_modes = PODI.choice_n_modes_tol(ev_list, tol)
            n_modes_list.append(n_modes)

            local_basis = podi.get_basis()
            local_pod_bases.append(local_basis)

        # Function that choose the 'n_bases' closest bases (closest value of x^(1))
        # and use 'interp_subspaces' to estimate a new basis
        n_bases = 10

        def choose_local_bases(local_pod_bases, n_bases, modes_list, xt1, xv1):
            import numpy as np
            from smt.applications import PODI

            interpolated_bases = []
            keep_index_list = []
            max_modes_list = []

            for value in xv1:
                sorted_ind = sorted(
                    range(xt1.shape[0]), key=lambda k: abs(xt1[:, 0] - value)[k]
                )
                keep_index = sorted_ind[:n_bases]
                keep_index_list.append(keep_index)
                input_matrices = []
                keep_xt1 = []
                max_modes = max(modes_list[keep_index])
                max_modes_list.append(max_modes)
                for i in keep_index:
                    input_matrices.append(local_pod_bases[i][:, :max_modes])
                    keep_xt1.append(xt1[i, 0])
                basis = PODI.interp_subspaces(
                    xt1=np.atleast_2d(keep_xt1).T,
                    input_matrices=input_matrices,
                    xn1=np.atleast_2d(value),
                    frechet=True,
                    print_global=False,
                )
                interpolated_bases.append(basis[0])
            return interpolated_bases, keep_index_list

        interpolated_bases, keep_index_list = choose_local_bases(
            local_pod_bases,
            n_bases=n_bases,
            modes_list=np.array(n_modes_list),
            xt1=xt1,
            xv1=xv1,
        )

        # Choosing a value from the validation inputs
        i = 0

        podi = PODI()
        j = []
        for ind in keep_index_list[i]:
            j += list(range(ind * nt2, (ind + 1) * nt2))
        podi.compute_pod(
            database=dbt[:, j], pod_type="local", local_basis=interpolated_bases[i]
        )
        n_modes = podi.get_n_modes()
        print(f"{n_modes} modes were kept.")

        # Choosing the default interp options
        # Setting the training values
        podi.set_training_values(xt=np.atleast_2d(xt[j]))

        # Training the models
        podi.train()

        # predicting the desired values with inputs
        values = podi.predict_values(np.atleast_2d(xv[i]))

        diff = dbv[:, i] - values[:, 0]

        rms_error = np.sqrt(np.mean(diff**2))
        plt.figure(figsize=(8, 5))
        plt.scatter(
            y,
            values,
            color="r",
            marker="x",
            s=15,
            alpha=1.0,
            label="prediction (mean)",
        )
        plt.scatter(
            y,
            dbv[:, i],
            color="b",
            marker="*",
            s=5,
            alpha=1.0,
            label="reference",
        )
        plt.plot([], [], color="w", label="rmse = " + str(round(rms_error, 5)))

        ax = plt.gca()
        ax.axes.xaxis.set_visible(False)

        plt.ylabel("u(x = " + str(xv[i, 0])[:4] + ")")
        plt.title(f"Estimation of u at x = ({str(xv[i, 0])[:4]}, {str(xv[i, 1])[:4]})")
        plt.legend()
        plt.show()


if __name__ == "__main__":
    unittest.main()



"""
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.
"""

import sys
from bisect import bisect_left

import numpy as np

import scipy
from packaging.version import Version

# Since scipy 1.15, derivative function has moved from scipy.misc to scipy.differentiate
# As derivative is used by several benchmarking problems, we initialize a constant
# once here as a proxy of the derivative function wrt the installed scipy version
if Version(scipy.__version__) >= Version("1.15"):
    SCIPY_DERIVATIVE = scipy.differentiate.derivative
else:
    SCIPY_DERIVATIVE = scipy.misc.derivative


def standardization(X, y):
    """

    We substract the mean from each variable. Then, we divide the values of each
    variable by its standard deviation. If scale_X_to_unit, we scale the input
    space X to the unit hypercube [0,1]^dim with dim the input dimension.

    Parameters
    ----------

    X: np.ndarray [n_obs, dim]
            - The input variables.

    y: np.ndarray [n_obs, 1]
            - The output variable.

    Returns
    -------

    X: np.ndarray [n_obs, dim]
          The standardized input matrix.

    y: np.ndarray [n_obs, 1]
          The standardized output vector.

    X_offset: list(dim)
            The mean (or the min if scale_X_to_unit=True) of each input variable.

    y_mean: list(1)
            The mean of the output variable.

    X_scale:  list(dim)
            The standard deviation of each input variable.

    y_std:  list(1)
            The standard deviation of the output variable.

    """

    X_offset = np.mean(X, axis=0)
    X_scale = X.std(axis=0, ddof=1)

    X_scale[np.abs(X_scale) < (100.0 * sys.float_info.epsilon)] = 1.0
    y_mean = np.mean(y, axis=0)
    y_std = y.std(axis=0, ddof=1)
    y_std[y_std == 0.0] = 1.0

    # scale X and y
    X = (X - X_offset) / X_scale
    y = (y - y_mean) / y_std
    return X, y, X_offset, y_mean, X_scale, y_std


def compute_relative_error(sm, xe=None, ye=None, kx=None):
    """
    Returns a normalized RMS error of the training points or the given points.

    Arguments
    ---------
    sm : Surrogate
        Surrogate model instance.
    xe : np.ndarray[ne, dim] or None
        Input values. If None, the input values at the training points are used instead.
    ye : np.ndarray[ne, 1] or None
        Output / deriv. values. If None, the training pt. outputs / derivs. are used.
    kx : int or None
        If None, we are checking the output values.
        If int, we are checking the derivs. w.r.t. the kx^{th} input variable (0-based).
    """

    if xe is not None and ye is not None:
        ye = ye.reshape((xe.shape[0], 1))
        if kx is None:
            ye2 = sm.predict_values(xe)
        else:
            ye2 = sm.predict_derivatives(xe, kx)
        return np.linalg.norm(ye2 - ye) / np.linalg.norm(ye)
    elif xe is None and ye is None:
        num = 0.0
        den = 0.0
        if kx is None:
            kx2 = 0
        else:
            kx2 += 1
        if kx2 not in sm.training_points[None]:
            raise ValueError(
                "There is no training point data available for kx %s" % kx2
            )
        xt, yt = sm.training_points[None][kx2]
        if kx is None:
            yt2 = sm.predict_values(xt)
        else:
            yt2 = sm.predict_derivatives(xt, kx)
        num = np.linalg.norm(yt2 - yt)
        den = np.linalg.norm(yt)
        return num / den


def compute_pva(sm, xe, ye):
    """
    Compute the Predictive Variance Adequacy (PVA) for a surrogate model.

    Parameters:
    - sm: The surrogate model object, expected to have `predict_values` and `predict_variances` methods.
    - xe: Input data for evaluation (N x d array).
    - ye: True output values (N x 1 array or equivalent).

    Returns:
    - pva: Predictive Variance Adequacy score (float).
    """
    ye = ye.reshape((xe.shape[0], 1))  # Ensure `ye` is column vector
    Nb = len(ye)  # Number of data points

    # Predicted values and variances
    ye_pred = sm.predict_values(xe)  # Predicted values (N x 1 array)
    variance = sm.predict_variances(xe)  # Predicted variances (N x 1 array)

    # Calculate squared error normalized by variance
    error = ((ye_pred - ye) ** 2) / variance

    # Compute PVA with logarithm
    pva = np.abs(np.log(np.sum(error) / Nb))

    return pva


def compute_rmse(sm, xe, ye):
    """
    Compute the Root Mean Square Error (RMSE) for a surrogate model.

    Parameters:
    - sm: The surrogate model object, expected to have a `predict_values` method.
    - xe: Input data for evaluation (N x d array).
    - ye: True output values (N x 1 array or equivalent).

    Returns:
    - rmse: Root Mean Square Error (float).
    """
    ye = ye.reshape((xe.shape[0], 1))  # Ensure `ye` is a column vector

    # Predicted values
    ye_pred = sm.predict_values(xe)  # Predicted values (N x 1 array)

    # Compute RMSE
    mse = np.mean((ye_pred - ye) ** 2)  # Mean Squared Error
    rmse = np.sqrt(mse)  # Root Mean Squared Error

    return rmse


def compute_q2(sm, xe, ye):
    """
    Compute the Q^2 validation criterion for a surrogate model.

    Parameters:
    - sm: The surrogate model object, expected to have a `predict_values` method.
    - xe: Input data for evaluation (N x d array).
    - ye: True output values (N x 1 array or equivalent).

    Returns:
    - Q2: Predictive coefficient of determination (float).
    """
    ye = ye.reshape((xe.shape[0], 1))  # Ensure `ye` is a column vector

    # Predicted values
    ye_pred = sm.predict_values(xe)  # Predicted values (N x 1 array)

    # Mean of true output values
    ye_mean = np.mean(ye)

    # Residual Sum of Squares (RSS) and Total Sum of Squares (TSS)
    rss = np.sum((ye - ye_pred) ** 2)  # Residual sum of squares
    tss = np.sum((ye - ye_mean) ** 2)  # Total sum of squares

    # Compute Q^2
    Q2 = 1 - (rss / tss)

    return Q2


def take_closest_number(myList, myNumber):
    """
    Assumes myList is sorted. Returns closest value to myNumber.

    If two numbers are equally close, return the smallest number.
    """
    pos = bisect_left(myList, myNumber)
    if pos == 0:
        return myList[0]
    if pos == len(myList):
        return myList[-1]
    before = myList[pos - 1]
    after = myList[pos]
    if after - myNumber < myNumber - before:
        return after
    else:
        return before


def take_closest_in_list(myList, x):
    vfunc = np.vectorize(take_closest_number, excluded=["myList"])
    return vfunc(myList=myList, myNumber=x)



"""
Author: Lisa Pretsch <<lisa.pretsch@tum.de>>

This package is distributed under New BSD license.
"""

import unittest

try:
    import matplotlib

    matplotlib.use("Agg")
    NO_MATPLOTLIB = False
except ImportError:
    NO_MATPLOTLIB = True

from smt.utils.sm_test_case import SMTestCase


class TestCCKRG(SMTestCase):
    @staticmethod
    def run_cckrg_example():
        import numpy as np
        import random

        from smt.problems import TensorProduct
        from smt.sampling_methods import LHS

        from smt.applications import CoopCompKRG

        # The problem is the exponential problem with dimension 10
        ndim = 10
        prob = TensorProduct(ndim=ndim, func="exp")

        # Example with three random components
        # (use physical components if available)
        ncomp = 3

        # Initial sampling
        samp = LHS(xlimits=prob.xlimits, random_state=42)
        np.random.seed(0)
        xt = samp(50)
        yt = prob(xt)
        np.random.seed(1)

        # Random design variable to component allocation
        comps = [*range(ncomp)]
        vars = [*range(ndim)]
        random.shuffle(vars)
        comp_var = np.full((ndim, ncomp), False)
        for c in comps:
            comp_size = int(ndim / ncomp)
            start = c * comp_size
            end = (c + 1) * comp_size
            if c + 1 == ncomp:
                end = max((c + 1) * comp_size, ndim)
            comp_var[vars[start:end], c] = True

        # Cooperative components Kriging model fit
        sm = CoopCompKRG()
        for active_coop_comp in comps:
            sm.set_training_values(xt, yt)
            sm.train(active_coop_comp, comp_var)

        # Prediction as for ordinary Kriging
        xpoint = (-5 + np.array([[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]])) / 10.0
        print(sm.predict_values(xpoint))
        print(sm.predict_variances(xpoint))

    # run scripts are used in documentation as documentation is not always rebuild
    # make a test run by pytest to test the run scripts
    @unittest.skipIf(NO_MATPLOTLIB, "Matplotlib not installed")
    def test_run_cckrg_example(self):
        self.run_cckrg_example()


if __name__ == "__main__":
    unittest.main()



import pickle
import zlib


def save(self, filename):
    serialized_data = pickle.dumps(self, protocol=5)
    compressed_data = zlib.compress(serialized_data)
    with open(filename, "wb") as file:
        file.write(compressed_data)


def load(filename):
    with open(filename, "rb") as file:
        compressed_data = file.read()

    serialized_data = zlib.decompress(compressed_data)
    sm = pickle.loads(serialized_data)

    return sm



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct  6 14:47:08 2023

@author: rcharayr
"""

import unittest

import numpy as np

try:
    import matplotlib

    matplotlib.use("Agg")
    NO_MATPLOTLIB = False
except ImportError:
    NO_MATPLOTLIB = True

import numpy.linalg as npl

from smt.applications import NestedLHS
from smt.applications.mfk import MFK
from smt.applications.mfkpls import MFKPLS
from smt.applications.mixed_integer import (
    MixedIntegerSamplingMethod,
)
from smt.sampling_methods import LHS
from smt.surrogate_models import (
    KPLS,
    KRG,
    MixIntKernelType,
)
from smt.design_space import (
    DesignSpace,
    CategoricalVariable,
    FloatVariable,
    IntegerVariable,
)


class TestMFKmixed(unittest.TestCase):
    # useful functions (g and h)
    def g_ZDT(self, x):
        x1 = x[1]
        x2 = x[2]
        x3 = x[3]
        x4 = x[4]
        if x[5] == 0:  # 0 -> blue
            x5 = 0.1
        if x[5] == 1:  # 1 -> red
            x5 = 0.5
        if x[5] == 2:  # 2 -> green
            x5 = 0.75

        out = 1 + (9 / (6 - 1)) * (x1 + x2 + x3 + x4 + x5)
        return out

    def g_DTLZ5(self, x):
        x1 = x[1]
        x2 = x[2]
        x3 = x[3]
        x4 = x[4]
        if x[5] == 0:  # 0 -> blue
            x5 = 0.1
        if x[5] == 1:  # 1 -> red
            x5 = 0.5
        if x[5] == 2:  # 2 -> green
            x5 = 0.75

        out = (
            (x1 - 0.5) ** 2
            + (x2 - 0.5) ** 2
            + (x3 - 0.5) ** 2
            + (x4 - 0.5) ** 2
            + (x5 - 0.5) ** 2
        )
        return out

    def h_ZDT1(self, x):
        x0 = x[0]
        out = 1 - np.sqrt(x0 / self.g_ZDT(x))
        return out

    def h_ZDT2(self, x):
        x0 = x[0]
        out = 1 - (x0 / self.g_ZDT(x)) ** 2
        return out

    def h_ZDT3(self, x):
        x0 = x[0]
        out = (
            1
            - np.sqrt(x0 / self.g_ZDT(x))
            - (x0 / self.g_ZDT(x)) * np.sin(10 * np.pi * x0)
        )
        return out

    # grouped functions
    def ZDT1_HF_mixt(self, x):
        x0 = x[0]
        y1 = x0
        y2 = self.g_ZDT(x) * self.h_ZDT1(x)

        fail = False
        return [y1, y2], fail

    def ZDT1_LF_mixt(self, x):
        x0 = x[0]
        y1 = 0.9 * x0 + 0.1
        y2 = (0.8 * self.g_ZDT(x) - 0.2) * (1.2 * self.h_ZDT1(x) + 0.2)

        fail = False
        return [y1, y2], fail

    ###############################################################################

    def ZDT2_HF_mixt(self, x):
        x0 = x[0]
        y1 = x0
        y2 = self.g_ZDT(x) * self.h_ZDT2(x)

        fail = False
        return [y1, y2], fail

    def ZDT2_LF_mixt(self, x):
        x0 = x[0]
        y1 = 0.8 * x0 + 0.2
        y2 = (0.9 * self.g_ZDT(x) + 0.2) * (1.1 * self.h_ZDT2(x) - 0.2)

        fail = False
        return [y1, y2], fail

    ###############################################################################

    def ZDT3_HF_mixt(self, x):
        x0 = x[0]
        y1 = x0
        y2 = self.g_ZDT(x) * self.h_ZDT3(x)
        fail = False
        return [y1, y2], fail

    def ZDT3_LF_mixt(self, x):
        x0 = x[0]
        y1 = 0.75 * x0 + 0.25
        y2 = self.g_ZDT(x) * (1.25 * self.h_ZDT3(x) - 0.25)
        fail = False
        return [y1, y2], fail

    ###############################################################################

    def DTLZ5_HF_mixt(self, x):
        x0 = x[0]
        y1 = (1 + self.g_DTLZ5(x)) * np.cos(0.5 * np.pi * x0)
        y2 = (1 + self.g_DTLZ5(x)) * np.sin(0.5 * np.pi * x0)
        fail = False
        return [y1, y2], fail

    def DTLZ5_LF_mixt(self, x):
        x0 = x[0]
        y1 = (1 + 0.8 * self.g_DTLZ5(x)) * np.cos(0.5 * np.pi * x0)
        y2 = (1 + 1.1 * self.g_DTLZ5(x)) * np.sin(0.5 * np.pi * x0)

        fail = False
        return [y1, y2], fail

    ###############################################################################

    def run_mfk_mixed_example(self):
        import matplotlib.pyplot as plt

        # KRG_METHODS = ["krg", "kpls", "mfk", "mfkpls"]
        # KRG_METHODS = ["krg"]
        # KRG_METHODS = ["kpls"]
        KRG_METHODS = ["mfk"]
        # KRG_METHODS = ["mfkpls"]

        pbm = "ZDT1"

        if pbm == "ZDT1":
            HF_fun = self.ZDT1_HF_mixt
            LF_fun = self.ZDT1_LF_mixt
        elif pbm == "ZDT2":
            HF_fun = self.ZDT2_HF_mixt
            LF_fun = self.ZDT2_LF_mixt
        elif pbm == "ZDT3":
            HF_fun = self.ZDT3_HF_mixt
            LF_fun = self.ZDT3_LF_mixt
        elif pbm == "DTLZ5":
            HF_fun = self.DTLZ5_HF_mixt
            LF_fun = self.DTLZ5_LF_mixt

        # ------------------------------------------------------------------------------

        # Instantiate the design space with all its design variables:
        ds = DesignSpace(
            [
                FloatVariable(0, 1),  # x0 continuous between 0 and 1
                FloatVariable(0, 1),  # x0 continuous between 0 and 1
                FloatVariable(0, 1),  # x0 continuous between 0 and 1
                FloatVariable(0, 1),  # x0 continuous between 0 and 1
                # OrdinalVariable(['0', '1']),  # x4 ordinal: 0 and 1; order is relevant
                IntegerVariable(0, 1),  # x4 integer between 0 and 1
                CategoricalVariable(
                    ["blue", "red", "green"]
                ),  # x5 categorical: blue, red or green; order is not relevant
            ]
        )

        # ------------------------------------------------------------------------------

        # Validation data:
        n_valid = 100  # validation set size
        ds.seed = 42
        samp = MixedIntegerSamplingMethod(
            LHS, ds, criterion="ese", random_state=ds.seed
        )
        x_valid, is_acting_valid = samp(n_valid, return_is_acting=True)

        y1_valid = np.zeros(n_valid)  # obj 1
        y2_valid = np.zeros(n_valid)  # obj 2
        y1_valid_LF = np.zeros(n_valid)  # obj 1
        y2_valid_LF = np.zeros(n_valid)  # obj 2
        for i in range(n_valid):
            res = HF_fun(x_valid[i, :])
            y1_valid[i] = res[0][0]  # obj 1
            y2_valid[i] = res[0][1]  # obj 2
            res_LF = LF_fun(x_valid[i, :])
            y1_valid_LF[i] = res_LF[0][0]  # obj 1
            y2_valid_LF[i] = res_LF[0][1]  # obj 2
        # ------------------------------------------------------------------------------

        # HF training data:
        n_train_HF = 20  # training set size
        n_train_LF = 40
        nlhs = NestedLHS(nlevel=2, design_space=ds)
        xt_LF, xt_HF = nlhs(n_train_HF)

        y1t_HF = np.zeros(n_train_HF)  # obj 1
        y2t_HF = np.zeros(n_train_HF)  # obj 2
        for i in range(n_train_HF):
            res = HF_fun(xt_HF[i, :])
            y1t_HF[i] = res[0][0]  # obj 1
            y2t_HF[i] = res[0][1]  # obj 2

        # ------------------------------------------------------------------------------

        for krg_method in KRG_METHODS:
            if "mfk" in krg_method:  # if multifi compute LF validation data
                y1t_LF = np.zeros(n_train_LF)  # obj 1
                y2t_LF = np.zeros(n_train_LF)  # obj 2
                for i in range(n_train_LF):
                    res_LF = LF_fun(xt_LF[i, :])
                    y1t_LF[i] = res_LF[0][0]  # obj 1
                    y2t_LF[i] = res_LF[0][1]  # obj 2

            if krg_method == "krg":
                print("KRG")
                ###########################
                # Mono-fidelity KRG
                ###########################

                sm1 = KRG(
                    design_space=ds,
                    theta0=[1e-2],
                    print_prediction=False,
                    corr="squar_exp",
                    categorical_kernel=MixIntKernelType.CONT_RELAX,
                )
                sm2 = KRG(
                    design_space=ds,
                    theta0=[1e-2],
                    print_prediction=False,
                    corr="squar_exp",
                    categorical_kernel=MixIntKernelType.CONT_RELAX,
                )

                # Set training data
                # obj 1
                sm1.set_training_values(xt_HF, y1t_HF)
                # obj 2
                sm2.set_training_values(xt_HF, y2t_HF)

            if krg_method == "kpls":
                print("KPLS")
                ###########################
                # Mono-fidelity KPLS
                ###########################

                sm1 = KPLS(
                    n_comp=3,
                    design_space=ds,
                    theta0=[1e-2],
                    print_prediction=False,
                    corr="squar_exp",
                    categorical_kernel=MixIntKernelType.CONT_RELAX,
                )
                sm2 = KPLS(
                    n_comp=3,
                    design_space=ds,
                    theta0=[1e-2],
                    print_prediction=False,
                    corr="squar_exp",
                    categorical_kernel=MixIntKernelType.CONT_RELAX,
                )

                # Set training data
                # obj 1
                sm1.set_training_values(xt_HF, y1t_HF)
                # obj 2
                sm2.set_training_values(xt_HF, y2t_HF)

            if krg_method == "mfk":
                print("MFK")
                ###########################
                # MFK
                ###########################

                sm1 = MFK(
                    design_space=ds,
                    theta0=[1e-2],
                    print_prediction=False,
                    corr="squar_exp",
                    categorical_kernel=MixIntKernelType.CONT_RELAX,
                    hyper_opt="Cobyla",
                )
                sm2 = MFK(
                    design_space=ds,
                    theta0=[1e-2],
                    print_prediction=False,
                    corr="squar_exp",
                    categorical_kernel=MixIntKernelType.CONT_RELAX,
                    hyper_opt="Cobyla",
                )

                # Set training data
                # obj 1
                sm1.set_training_values(xt_LF[:, :], y1t_LF, name=0)
                sm1.set_training_values(xt_HF[:, :], y1t_HF)
                # obj 2
                sm2.set_training_values(xt_LF[:, :], y2t_LF, name=0)
                sm2.set_training_values(xt_HF[:, :], y2t_HF)

            if krg_method == "mfkpls":
                print("MFKPLS")
                ###########################
                # MFKKPLS
                ##########################

                sm1 = MFKPLS(
                    n_comp=3,
                    design_space=ds,
                    theta0=[1e-2],
                    print_prediction=False,
                    corr="squar_exp",
                    categorical_kernel=MixIntKernelType.CONT_RELAX,
                    hyper_opt="Cobyla",
                )
                sm2 = MFKPLS(
                    n_comp=3,
                    design_space=ds,
                    theta0=[1e-2],
                    print_prediction=False,
                    corr="squar_exp",
                    categorical_kernel=MixIntKernelType.CONT_RELAX,
                    hyper_opt="Cobyla",
                )

                # Set training data
                # obj 1
                sm1.set_training_values(xt_LF[:, :], y1t_LF, name=0)
                sm1.set_training_values(xt_HF[:, :], y1t_HF)
                # obj 2
                sm2.set_training_values(xt_LF[:, :], y2t_LF, name=0)
                sm2.set_training_values(xt_HF[:, :], y2t_HF)

            # Train the models:
            import time

            t0 = time.time()
            sm1.train()
            t1 = time.time()
            sm2.train()
            t2 = time.time()
            print("t1-t0=", t1 - t0)
            print("t2-t1=", t2 - t1)

            # Compute errors:
            # obj 1
            y1_predict = np.ravel(sm1.predict_values(x_valid))
            if "mfk" in krg_method:
                y1_predict_LF = np.ravel(sm1._predict_intermediate_values(x_valid, 1))
                _y1_var_predict_LF = np.ravel(
                    sm1.predict_variances_all_levels(x_valid)[0]
                )
            y1_var_predict = np.ravel(sm1.predict_variances(x_valid))
            _y1_sd_predict = np.sqrt(y1_var_predict)
            # obj 2
            y2_predict = np.ravel(sm2.predict_values(x_valid))
            if "mfk" in krg_method:
                y2_predict_LF = np.ravel(sm2._predict_intermediate_values(x_valid, 1))
                _y2_var_predict_LF = np.ravel(
                    sm2.predict_variances_all_levels(x_valid)[0]
                )
            y2_var_predict = np.ravel(sm2.predict_variances(x_valid))
            _y2_sd_predict = np.sqrt(y2_var_predict)

            # obj 1:
            # print("y1_predict - y1_valid =", y1_predict - y1_valid)
            print("norme2(y1_predict - y1_valid) =", npl.norm(y1_predict - y1_valid, 2))
            if "mfk" in krg_method:
                print(
                    "norme2(y1_predict_LF - y1_valid_LF) =",
                    npl.norm(y1_predict_LF - y1_valid_LF, 2),
                )
            # print("y1_sd_predict =", y1_sd_predict)

            # obj 2:
            # print("y2_predict - y2_valid =", y2_predict - y2_valid)
            print("norme2(y2_predict - y2_valid) =", npl.norm(y2_predict - y2_valid, 2))
            if "mfk" in krg_method:
                print(
                    "norme2(y2_predict_LF - y2_valid_LF) =",
                    npl.norm(y2_predict_LF - y2_valid_LF, 2),
                )
            # print("y2_sd_predict =", y2_sd_predict)

            # PLOTS:
            plt.figure()
            plt.title("y1")
            plt.xlabel("y1_valid")
            plt.ylabel("y1_predict")
            plt.scatter(y1_valid, y1_predict)
            plt.plot(np.linspace(0, 1.5, 100), np.linspace(0, 1.5, 100), color="red")
            plt.show()

            plt.figure()
            plt.title("y2")
            plt.xlabel("y2_valid")
            plt.ylabel("y2_predict")
            plt.scatter(y2_valid, y2_predict)
            plt.plot(np.linspace(0, 9, 100), np.linspace(0, 9, 100), color="red")
            plt.show()

            if "mfk" in krg_method:
                plt.figure()
                plt.title("y1")
                plt.xlabel("y1_valid_LF")
                plt.ylabel("y1_predict_LF")
                plt.scatter(y1_valid_LF, y1_predict_LF)
                plt.plot(
                    np.linspace(0, 1.5, 100), np.linspace(0, 1.5, 100), color="red"
                )
                plt.show()

                plt.figure()
                plt.title("y2")
                plt.xlabel("y2_valid_LF")
                plt.ylabel("y2_predict_LF")
                plt.scatter(y2_valid_LF, y2_predict_LF)
                plt.plot(np.linspace(0, 9, 100), np.linspace(0, 9, 100), color="red")
                plt.show()

            # ------------------------------------------------------------------------------

            # PLOTS tests:
            plt.figure()
            plt.title("y1")
            plt.xlabel("y1t_HF")
            plt.ylabel("np.ravel(sm1.predict_values(xt_HF))")
            plt.scatter(y1t_HF, np.ravel(sm1.predict_values(xt_HF)))
            plt.plot(np.linspace(0, 1.5, 100), np.linspace(0, 1.5, 100), color="red")
            plt.show()

            plt.figure()
            plt.title("y2")
            plt.xlabel("y2t_HF")
            plt.ylabel("np.ravel(sm2.predict_values(xt_HF))")
            plt.scatter(y2t_HF, np.ravel(sm2.predict_values(xt_HF)))
            plt.plot(np.linspace(0, 9, 100), np.linspace(0, 9, 100), color="red")
            plt.show()

            if "mfk" in krg_method:
                plt.figure()
                plt.title("y1")
                plt.xlabel("y1t_LF")
                plt.ylabel("np.ravel(sm1._predict_intermediate_values(xt_LF,1))")
                plt.scatter(
                    y1t_LF, np.ravel(sm1._predict_intermediate_values(xt_LF, 1))
                )
                plt.plot(
                    np.linspace(0, 1.5, 100), np.linspace(0, 1.5, 100), color="red"
                )
                plt.show()

                plt.figure()
                plt.title("y2")
                plt.xlabel("y2t_LF")
                plt.ylabel("np.ravel(sm2._predict_intermediate_values(xt_LF,1))")
                plt.scatter(
                    y2t_LF, np.ravel(sm2._predict_intermediate_values(xt_LF, 1))
                )
                plt.plot(np.linspace(0, 9, 100), np.linspace(0, 9, 100), color="red")
                plt.show()
                # ------------------------------------------------------------------------------

    def run_mfkpls_mixed_example(self):
        import matplotlib.pyplot as plt

        # KRG_METHODS = ["krg", "kpls", "mfk", "mfkpls"]
        # KRG_METHODS = ["krg"]
        # KRG_METHODS = ["kpls"]
        # KRG_METHODS = ["mfk"]
        KRG_METHODS = ["mfkpls"]

        pbm = "ZDT1"

        if pbm == "ZDT1":
            HF_fun = self.ZDT1_HF_mixt
            LF_fun = self.ZDT1_LF_mixt
        elif pbm == "ZDT2":
            HF_fun = self.ZDT2_HF_mixt
            LF_fun = self.ZDT2_LF_mixt
        elif pbm == "ZDT3":
            HF_fun = self.ZDT3_HF_mixt
            LF_fun = self.ZDT3_LF_mixt
        elif pbm == "DTLZ5":
            HF_fun = self.DTLZ5_HF_mixt
            LF_fun = self.DTLZ5_LF_mixt

        # ------------------------------------------------------------------------------

        # Instantiate the design space with all its design variables:
        ds = DesignSpace(
            [
                FloatVariable(0, 1),  # x0 continuous between 0 and 1
                FloatVariable(0, 1),  # x0 continuous between 0 and 1
                FloatVariable(0, 1),  # x0 continuous between 0 and 1
                FloatVariable(0, 1),  # x0 continuous between 0 and 1
                # OrdinalVariable(['0', '1']),  # x4 ordinal: 0 and 1; order is relevant
                IntegerVariable(0, 1),  # x4 integer between 0 and 1
                CategoricalVariable(
                    ["blue", "red", "green"]
                ),  # x5 categorical: blue, red or green; order is not relevant
            ]
        )

        # ------------------------------------------------------------------------------

        # Validation data:
        n_valid = 100  # validation set size
        x_valid, is_acting_valid = ds.sample_valid_x(n_valid)

        y1_valid = np.zeros(n_valid)  # obj 1
        y2_valid = np.zeros(n_valid)  # obj 2
        y1_valid_LF = np.zeros(n_valid)  # obj 1
        y2_valid_LF = np.zeros(n_valid)  # obj 2
        for i in range(n_valid):
            res = HF_fun(x_valid[i, :])
            y1_valid[i] = res[0][0]  # obj 1
            y2_valid[i] = res[0][1]  # obj 2
            res_LF = LF_fun(x_valid[i, :])
            y1_valid_LF[i] = res_LF[0][0]  # obj 1
            y2_valid_LF[i] = res_LF[0][1]  # obj 2
        # ------------------------------------------------------------------------------

        # HF training data:
        n_train_HF = 20  # training set size
        xt_HF, is_acting_t_HF = ds.sample_valid_x(n_train_HF)

        y1t_HF = np.zeros(n_train_HF)  # obj 1
        y2t_HF = np.zeros(n_train_HF)  # obj 2
        for i in range(n_train_HF):
            res = HF_fun(xt_HF[i, :])
            y1t_HF[i] = res[0][0]  # obj 1
            y2t_HF[i] = res[0][1]  # obj 2

        # ------------------------------------------------------------------------------

        for krg_method in KRG_METHODS:
            if "mfk" in krg_method:  # if multifi compute LF validation data
                n_train_LF = 40  # training set size   IF MULTIFI
                sampling = MixedIntegerSamplingMethod(LHS, ds, criterion="ese")
                xt_LF = np.concatenate(
                    (xt_HF, sampling.expand_lhs(xt_HF, n_train_LF - n_train_HF)), axis=0
                )

                y1t_LF = np.zeros(n_train_LF)  # obj 1
                y2t_LF = np.zeros(n_train_LF)  # obj 2
                for i in range(n_train_LF):
                    res_LF = LF_fun(xt_LF[i, :])
                    y1t_LF[i] = res_LF[0][0]  # obj 1
                    y2t_LF[i] = res_LF[0][1]  # obj 2

            if krg_method == "krg":
                print("KRG")
                ###########################
                # Mono-fidelity KRG
                ###########################

                sm1 = KRG(
                    design_space=ds,
                    theta0=[1e-2],
                    print_prediction=False,
                    corr="squar_exp",
                    categorical_kernel=MixIntKernelType.CONT_RELAX,
                )
                sm2 = KRG(
                    design_space=ds,
                    theta0=[1e-2],
                    print_prediction=False,
                    corr="squar_exp",
                    categorical_kernel=MixIntKernelType.CONT_RELAX,
                )

                # Set training data
                # obj 1
                sm1.set_training_values(xt_HF, y1t_HF)
                # obj 2
                sm2.set_training_values(xt_HF, y2t_HF)

            if krg_method == "kpls":
                print("KPLS")
                ###########################
                # Mono-fidelity KPLS
                ###########################

                sm1 = KPLS(
                    n_comp=3,
                    design_space=ds,
                    theta0=[1e-2],
                    print_prediction=False,
                    corr="squar_exp",
                    categorical_kernel=MixIntKernelType.CONT_RELAX,
                )
                sm2 = KPLS(
                    n_comp=3,
                    design_space=ds,
                    theta0=[1e-2],
                    print_prediction=False,
                    corr="squar_exp",
                    categorical_kernel=MixIntKernelType.CONT_RELAX,
                )

                # Set training data
                # obj 1
                sm1.set_training_values(xt_HF, y1t_HF)
                # obj 2
                sm2.set_training_values(xt_HF, y2t_HF)

            if krg_method == "mfk":
                print("MFK")
                ###########################
                # MFK
                ###########################

                sm1 = MFK(
                    design_space=ds,
                    theta0=[1e-2],
                    print_prediction=False,
                    corr="squar_exp",
                    categorical_kernel=MixIntKernelType.CONT_RELAX,
                )
                sm2 = MFK(
                    design_space=ds,
                    theta0=[1e-2],
                    print_prediction=False,
                    corr="squar_exp",
                    categorical_kernel=MixIntKernelType.CONT_RELAX,
                )

                # Set training data
                # obj 1
                sm1.set_training_values(xt_LF[:, :], y1t_LF, name=0)
                sm1.set_training_values(xt_HF[:, :], y1t_HF)
                # obj 2
                sm2.set_training_values(xt_LF[:, :], y2t_LF, name=0)
                sm2.set_training_values(xt_HF[:, :], y2t_HF)

            if krg_method == "mfkpls":
                print("MFKPLS")
                ###########################
                # MFKKPLS
                ##########################

                sm1 = MFKPLS(
                    n_comp=3,
                    design_space=ds,
                    theta0=[1e-2],
                    print_prediction=False,
                    corr="squar_exp",
                    categorical_kernel=MixIntKernelType.CONT_RELAX,
                )
                sm2 = MFKPLS(
                    n_comp=3,
                    design_space=ds,
                    theta0=[1e-2],
                    print_prediction=False,
                    corr="squar_exp",
                    categorical_kernel=MixIntKernelType.CONT_RELAX,
                )

                # Set training data
                # obj 1
                sm1.set_training_values(xt_LF[:, :], y1t_LF, name=0)
                sm1.set_training_values(xt_HF[:, :], y1t_HF)
                # obj 2
                sm2.set_training_values(xt_LF[:, :], y2t_LF, name=0)
                sm2.set_training_values(xt_HF[:, :], y2t_HF)

            # Train the models:
            import time

            t0 = time.time()
            sm1.train()
            t1 = time.time()
            sm2.train()
            t2 = time.time()
            print("t1-t0=", t1 - t0)
            print("t2-t1=", t2 - t1)

            # Compute errors:
            # obj 1
            y1_predict = np.ravel(sm1.predict_values(x_valid))
            if "mfk" in krg_method:
                y1_predict_LF = np.ravel(sm1._predict_intermediate_values(x_valid, 1))
                _y1_var_predict_LF = np.ravel(
                    sm1.predict_variances_all_levels(x_valid)[0]
                )
            y1_var_predict = np.ravel(sm1.predict_variances(x_valid))
            _y1_sd_predict = np.sqrt(y1_var_predict)
            # obj 2
            y2_predict = np.ravel(sm2.predict_values(x_valid))
            if "mfk" in krg_method:
                y2_predict_LF = np.ravel(sm2._predict_intermediate_values(x_valid, 1))
                _y2_var_predict_LF = np.ravel(
                    sm2.predict_variances_all_levels(x_valid)[0]
                )
            y2_var_predict = np.ravel(sm2.predict_variances(x_valid))
            _y2_sd_predict = np.sqrt(y2_var_predict)

            # obj 1:
            # print("y1_predict - y1_valid =", y1_predict - y1_valid)
            print("norme2(y1_predict - y1_valid) =", npl.norm(y1_predict - y1_valid, 2))
            if "mfk" in krg_method:
                print(
                    "norme2(y1_predict_LF - y1_valid_LF) =",
                    npl.norm(y1_predict_LF - y1_valid_LF, 2),
                )
            # print("y1_sd_predict =", y1_sd_predict)

            # obj 2:
            # print("y2_predict - y2_valid =", y2_predict - y2_valid)
            print("norme2(y2_predict - y2_valid) =", npl.norm(y2_predict - y2_valid, 2))
            if "mfk" in krg_method:
                print(
                    "norme2(y2_predict_LF - y2_valid_LF) =",
                    npl.norm(y2_predict_LF - y2_valid_LF, 2),
                )
            # print("y2_sd_predict =", y2_sd_predict)

            # PLOTS:
            plt.figure()
            plt.title("y1")
            plt.xlabel("y1_valid")
            plt.ylabel("y1_predict")
            plt.scatter(y1_valid, y1_predict)
            plt.plot(np.linspace(0, 1.5, 100), np.linspace(0, 1.5, 100), color="red")
            plt.show()

            plt.figure()
            plt.title("y2")
            plt.xlabel("y2_valid")
            plt.ylabel("y2_predict")
            plt.scatter(y2_valid, y2_predict)
            plt.plot(np.linspace(0, 9, 100), np.linspace(0, 9, 100), color="red")
            plt.show()

            if "mfk" in krg_method:
                plt.figure()
                plt.title("y1")
                plt.xlabel("y1_valid_LF")
                plt.ylabel("y1_predict_LF")
                plt.scatter(y1_valid_LF, y1_predict_LF)
                plt.plot(
                    np.linspace(0, 1.5, 100), np.linspace(0, 1.5, 100), color="red"
                )
                plt.show()

                plt.figure()
                plt.title("y2")
                plt.xlabel("y2_valid_LF")
                plt.ylabel("y2_predict_LF")
                plt.scatter(y2_valid_LF, y2_predict_LF)
                plt.plot(np.linspace(0, 9, 100), np.linspace(0, 9, 100), color="red")
                plt.show()

            # ------------------------------------------------------------------------------

            # PLOTS tests:
            plt.figure()
            plt.title("y1")
            plt.xlabel("y1t_HF")
            plt.ylabel("np.ravel(sm1.predict_values(xt_HF))")
            plt.scatter(y1t_HF, np.ravel(sm1.predict_values(xt_HF)))
            plt.plot(np.linspace(0, 1.5, 100), np.linspace(0, 1.5, 100), color="red")
            plt.show()

            plt.figure()
            plt.title("y2")
            plt.xlabel("y2t_HF")
            plt.ylabel("np.ravel(sm2.predict_values(xt_HF))")
            plt.scatter(y2t_HF, np.ravel(sm2.predict_values(xt_HF)))
            plt.plot(np.linspace(0, 9, 100), np.linspace(0, 9, 100), color="red")
            plt.show()

            if "mfk" in krg_method:
                plt.figure()
                plt.title("y1")
                plt.xlabel("y1t_LF")
                plt.ylabel("np.ravel(sm1._predict_intermediate_values(xt_LF,1))")
                plt.scatter(
                    y1t_LF, np.ravel(sm1._predict_intermediate_values(xt_LF, 1))
                )
                plt.plot(
                    np.linspace(0, 1.5, 100), np.linspace(0, 1.5, 100), color="red"
                )
                plt.show()

                plt.figure()
                plt.title("y2")
                plt.xlabel("y2t_LF")
                plt.ylabel("np.ravel(sm2._predict_intermediate_values(xt_LF,1))")
                plt.scatter(
                    y2t_LF, np.ravel(sm2._predict_intermediate_values(xt_LF, 1))
                )
                plt.plot(np.linspace(0, 9, 100), np.linspace(0, 9, 100), color="red")
                plt.show()
                # ------------------------------------------------------------------------------

    # run scripts are used in documentation as documentation is not always rebuild
    # make a test run by pytest to test the run scripts
    @unittest.skipIf(NO_MATPLOTLIB, "Matplotlib not installed")
    def test_mfkpls_mixed(self):
        self.run_mfk_mixed_example()
        self.run_mfkpls_mixed_example()


if __name__ == "__main__":
    TestMFKmixed().run_mfk_mixed_example()
    TestMFKmixed().run_mfkpls_mixed_example()
    unittest.main()



# -*- coding: utf-8 -*-
"""
Created on Wed Oct  9 16:17:04 2024

@author: mcastano
"""

import unittest

import numpy as np

try:
    import matplotlib

    matplotlib.use("Agg")
    NO_MATPLOTLIB = False
except ImportError:
    NO_MATPLOTLIB = True

from copy import deepcopy

from smt.applications.mfck import MFCK
from smt.problems import TensorProduct
from smt.sampling_methods import FullFactorial
from smt.utils.silence import Silence
from smt.utils.sm_test_case import SMTestCase

print_output = False


class TestMFCK(SMTestCase):
    def setUp(self):
        self.nt = 100
        self.ne = 100
        self.ndim = 3

    def test_mfck(self):
        self.problems = ["exp"]  # , "tanh", "cos"]

        for fname in self.problems:
            prob = TensorProduct(ndim=self.ndim, func=fname)
            sampling = FullFactorial(xlimits=prob.xlimits, clip=True)

            np.random.seed(0)
            xt = sampling(self.nt)
            yt = prob(xt)
            for i in range(self.ndim):
                yt = np.concatenate((yt, prob(xt, kx=i)), axis=1)

            y_lf = 2 * prob(xt) + 2
            x_lf = deepcopy(xt)
            np.random.seed(1)

            sm = MFCK(hyper_opt="Cobyla")
            if sm.options.is_declared("xlimits"):
                sm.options["xlimits"] = prob.xlimits
            sm.options["print_global"] = False

            sm.set_training_values(xt, yt[:, 0])
            sm.set_training_values(x_lf, y_lf[:, 0], name=0)

            with Silence():
                sm.train()

            m, c = sm._predict(xt)

            num = np.linalg.norm(m[:, 0] - yt[:, 0])
            den = np.linalg.norm(yt[:, 0])

            t_error = num / den

            self.assert_error(t_error, 0.0, 1e-5, 1e-5)

    @staticmethod
    def run_mfck_example():
        import matplotlib.pyplot as plt
        import numpy as np

        from smt.applications.mfck import MFCK
        from smt.applications.mfk import NestedLHS

        # low fidelity model
        def lf_function(x):
            import numpy as np

            return (
                0.5 * ((x * 6 - 2) ** 2) * np.sin((x * 6 - 2) * 2)
                + (x - 0.5) * 10.0
                - 5
            )

        # high fidelity model
        def hf_function(x):
            import numpy as np

            return ((x * 6 - 2) ** 2) * np.sin((x * 6 - 2) * 2)

        # Problem set up
        xlimits = np.array([[0.0, 1.0]])
        xdoes = NestedLHS(nlevel=2, xlimits=xlimits, random_state=0)
        xt_c, xt_e = xdoes(5)

        # Delta value for the non-nested difference applied in the LF
        delta = 0.05
        rnd_state = 1

        np.random.seed(rnd_state)
        deltas = np.random.uniform(-delta, delta, np.shape(xt_c))
        x_LF = xt_c + deltas
        x_LF = np.clip(x_LF, xlimits[0][0], xlimits[0][1])

        # Evaluate the HF and LF functions
        yt_e = hf_function(xt_e)
        yt_c = lf_function(x_LF)

        sm_non_nested = MFCK(theta0=xt_e.shape[1] * [0.5], corr="squar_exp")

        # low-fidelity dataset names being integers from 0 to level-1
        sm_non_nested.set_training_values(x_LF, yt_c, name=0)
        # high-fidelity dataset without name
        sm_non_nested.set_training_values(xt_e, yt_e)

        # train the model
        sm_non_nested.train()

        x = np.linspace(0, 1, 101, endpoint=True).reshape(-1, 1)

        m_non_nested, c_non_nested = sm_non_nested.predict_all_levels(x)

        plt.figure()

        plt.plot(x, hf_function(x), label="reference HF")
        plt.plot(x, lf_function(x), label="reference LF")
        plt.plot(x, m_non_nested[1], linestyle="-.", label="mean_gp_non_nested")
        plt.scatter(xt_e, yt_e, marker="o", color="k", label="HF doe")
        plt.scatter(x_LF, yt_c, marker="*", color="c", label="LF non-nested doe")

        plt.legend(loc=0)
        plt.ylim(-10, 17)
        plt.xlim(-0.1, 1.1)
        plt.xlabel(r"$x$")
        plt.ylabel(r"$y$")

        plt.show()

    # run scripts are used in documentation as documentation is not always rebuild
    # make a test run by pytest to test the run scripts
    @unittest.skipIf(NO_MATPLOTLIB, "Matplotlib not installed")
    def test_run_mfck_example(self):
        self.run_mfck_example()


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
Created on Mon Apr 27 15:36:13 2020

@author: Vincent Drouet and Nathalie Bartoli
in order to validate the variance formula for multifidelity on the branin 2D function
Comparisons are based  on the paper Le Gratiet et Cannamela 2015 :

Le Gratiet, L., & Cannamela, C. (2015).
Cokriging-based sequential design strategies using fast cross-validation
techniques for multi-fidelity computer codes. Technometrics, 57(3), 418-427.
https://doi.org/10.1080/00401706.2014.928233
"""

import unittest

import numpy as np

from smt.applications.mfk import MFK, NestedLHS
from smt.sampling_methods import LHS
from smt.utils.sm_test_case import SMTestCase

print_output = True

# %%
# Define the low and high fidelity models
# Example on a 2D problem: Branin  Function
A_corr = 0


# high fidelity model
def HF(point):
    # Expensive Forretal function
    res = (
        (
            point[:, 1]
            - (5.1 / (4 * np.pi**2)) * point[:, 0] ** 2
            + (5 / np.pi) * point[:, 0]
            - 6
        )
        ** 2
        + 10 * (1 - (1 / 8 / np.pi)) * np.cos(point[:, 0])
        + 10
        + 5 * point[:, 0]
    )
    return res


# low fidelity model
def LF(point):
    # Cheap Forretal function
    res = (
        HF(point)
        - (0.5 * A_corr**2 + A_corr + 0.2)
        * (
            point[:, 1]
            - 5.1 / 4 / np.pi**2 * point[:, 0] ** 2
            + 5.0 / np.pi * point[:, 0]
            - 6
        )
        ** 2
    )
    return res


class TestMFK_variance(SMTestCase):
    @staticmethod
    def corr(x1, x2, sm, lvl):
        # Normalization of data with   X_scale and X_offset from sm
        # lvl start to 1
        X_scale = sm.X_scale
        X_offset = sm.X_offset
        X1 = (x1 - X_offset) / X_scale
        X2 = (x2 - X_offset) / X_scale
        thetas = sm.optimal_theta[lvl - 1]
        if sm.options["corr"] == "abs_exp":
            p = 1
        else:
            p = 2
        return np.prod(np.exp(-thetas * np.abs(X1 - X2) ** p))

    @staticmethod
    def hyperparam_LG(sm):
        # computation of the hyperparameters   based on the formula from LeGratiet 2015

        y_t_lf = sm.training_points[0][0][1]
        n_LF = y_t_lf.shape[0]
        C = sm.optimal_par[0]["C"]
        R = C @ C.T
        H = np.ones((n_LF, 1))
        R_inv_H = np.linalg.solve(R, H)
        M = H.T @ R_inv_H
        M_inv = np.linalg.inv(M)
        R_inv_y = np.linalg.solve(R, y_t_lf)
        beta_LG_1 = M_inv @ H.T @ R_inv_y
        sigma2_LG_1 = (
            (y_t_lf - H @ beta_LG_1).T
            @ np.linalg.solve(R, y_t_lf - H @ beta_LG_1)
            / (n_LF - 1)
        )[0, 0]

        y_t_hf = sm.training_points[None][0][1]
        n_HF = y_t_hf.shape[0]
        y_D_l = y_t_lf[-n_HF:, :].reshape(
            -1, 1
        )  # to get y^l-1(D^l) we need n_HF last points of y_t_lf
        H = np.hstack((y_D_l, np.ones((n_HF, 1))))
        C = sm.optimal_par[1]["C"]
        R = C @ C.T
        R_inv_H = np.linalg.solve(R, H)
        M = H.T @ R_inv_H
        M_inv = np.linalg.inv(M)
        R_inv_y = np.linalg.solve(R, y_t_hf)
        rho_beta = M_inv @ H.T @ R_inv_y
        rho_LG = rho_beta[0, 0]
        beta_LG_2 = rho_beta[1, 0]
        sigma2_LG_2 = (
            (y_t_hf - H @ rho_beta).T
            @ np.linalg.solve(R, y_t_hf - H @ rho_beta)
            / (n_HF - 2)
        )[0, 0]
        sigma2_rho_LG = rho_LG**2 + sigma2_LG_2 * M_inv[0, 0]

        return beta_LG_1, sigma2_LG_1, beta_LG_2, sigma2_LG_2, rho_LG, sigma2_rho_LG

    @staticmethod
    def Mu_LG_LG(x, sm):
        # computation of the mean based on the formula from LeGratiet 2015
        # and using beta from LeGratiet 2015
        (
            beta_LG_1,
            sigma2_LG_1,
            beta_LG_2,
            sigma2_LG_2,
            rho_LG,
            sigma2_rho_LG,
        ) = TestMFK_variance.hyperparam_LG(sm)
        C = sm.optimal_par[0]["C"]
        R = C @ C.T
        x_t_lf = sm.training_points[0][0][0]
        y_t_lf = sm.training_points[0][0][1]
        n_LF = x_t_lf.shape[0]
        r_x = np.empty((n_LF, 1))
        for i in range(n_LF):
            r_x[i, 0] = TestMFK_variance.corr(x, x_t_lf[i], sm, 1)
        M = np.linalg.solve(R, y_t_lf - beta_LG_1 * np.ones((n_LF, 1)))
        mu_0 = beta_LG_1 + r_x.T @ M

        x_t_hf = sm.training_points[None][0][0]
        y_t_hf = sm.training_points[None][0][1]
        n_HF = x_t_hf.shape[0]
        C = sm.optimal_par[1]["C"]
        R = C @ C.T
        r_x = np.empty((n_HF, 1))
        for i in range(n_HF):
            r_x[i, 0] = TestMFK_variance.corr(x, x_t_hf[i], sm, 2)
        M = np.linalg.solve(
            R,
            y_t_hf
            - beta_LG_2 * np.ones((n_HF, 1))
            - rho_LG * y_t_lf[-n_HF:, 0, np.newaxis],
        )
        return rho_LG * mu_0 + beta_LG_2 + r_x.T @ M

    @staticmethod
    def Mu_LG_sm(x, sm):
        # computation of the mean based on the formula from LeGratiet 2015
        # and using beta from sm
        beta = sm.optimal_par[0]["beta"]
        C = sm.optimal_par[0]["C"]
        R = C @ C.T
        x_t_lf = sm.training_points[0][0][0]
        y_t_lf = sm.training_points[0][0][1]
        n_LF = x_t_lf.shape[0]
        r_x = np.empty((n_LF, 1))
        for i in range(n_LF):
            r_x[i, 0] = TestMFK_variance.corr(x, x_t_lf[i], sm, 1)
        M = np.linalg.solve(R, y_t_lf - beta * np.ones((n_LF, 1)))
        mu_0 = beta + r_x.T @ M

        x_t_hf = sm.training_points[None][0][0]
        y_t_hf = sm.training_points[None][0][1]
        n_HF = x_t_hf.shape[0]
        beta = sm.optimal_par[1]["beta"]
        rho = beta[0]
        C = sm.optimal_par[1]["C"]
        R = C @ C.T
        r_x = np.empty((n_HF, 1))
        for i in range(n_HF):
            r_x[i, 0] = TestMFK_variance.corr(x, x_t_hf[i], sm, 2)
        M = np.linalg.solve(
            R,
            y_t_hf - beta[1] * np.ones((n_HF, 1)) - rho * y_t_lf[-n_HF:, 0, np.newaxis],
        )
        return rho * mu_0 + beta[1] + r_x.T @ M

    @staticmethod
    def Cov_LG_sm(x1, x2, sm):
        # computation of the covariance  based on the formula from LeGratiet 2015
        # Using the Sigma2 output from sm
        x_t_lf = sm.training_points[0][0][0]
        y_t_lf = sm.training_points[0][0][1]
        n_LF = x_t_lf.shape[0]
        sigma2 = sm.optimal_par[0]["sigma2"]
        C = sm.optimal_par[0]["C"]
        R = C @ C.T
        r_x1 = np.empty((n_LF, 1))
        r_x2 = np.empty((n_LF, 1))
        for i in range(n_LF):
            r_x1[i, 0] = TestMFK_variance.corr(x1, x_t_lf[i], sm, 1)
            r_x2[i, 0] = TestMFK_variance.corr(x2, x_t_lf[i], sm, 1)
        h_x1 = np.ones((1, 1))
        h_x2 = np.ones((1, 1))
        H = np.ones((n_LF, 1))
        hr_1 = np.vstack((h_x1, r_x1))
        hr_2 = np.vstack((h_x2, r_x2))
        M1 = np.hstack((np.zeros((1, 1)), H.T))
        M2 = np.hstack((H, R))
        M3 = np.vstack((M1, M2))
        M4 = np.linalg.solve(M3, hr_2)  # M4 = M3^-1 @ hr_2
        k_0 = sigma2 * (
            TestMFK_variance.corr(x1, x2, sm, 1) - hr_1.T @ M4
        )  # covariance of level 0

        x_t_hf = sm.training_points[None][0][0]
        n_HF = x_t_hf.shape[0]
        sigma2 = sm.optimal_par[1]["sigma2"]
        (var_all_pred, sigma2_rho) = sm.predict_variances_all_levels(x1)
        sigma2_rho = sigma2_rho[0]
        C = sm.optimal_par[1]["C"]
        R = C @ C.T
        r_x1 = np.empty((n_HF, 1))
        r_x2 = np.empty((n_HF, 1))
        for i in range(n_HF):
            r_x1[i, 0] = TestMFK_variance.corr(x1, x_t_hf[i], sm, 2)
            r_x2[i, 0] = TestMFK_variance.corr(x2, x_t_hf[i], sm, 2)
        mu_x1 = sm._predict_intermediate_values(x1, 1).reshape(-1, 1)
        mu_x2 = sm._predict_intermediate_values(x2, 1).reshape(-1, 1)
        h_x1 = np.vstack((mu_x1, np.ones((1, 1))))
        h_x2 = np.vstack((mu_x2, np.ones((1, 1))))
        y_D_l = y_t_lf[-n_HF:, :].reshape(
            -1, 1
        )  # to get y^l-1(D^l) we need n_HF last points of y_t_lf
        H = np.hstack((y_D_l, np.ones((n_HF, 1))))
        hr_1 = np.vstack((h_x1, r_x1))
        hr_2 = np.vstack((h_x2, r_x2))
        M1 = np.hstack((np.zeros((2, 2)), H.T))
        M2 = np.hstack((H, R))
        M3 = np.vstack((M1, M2))
        M4 = np.linalg.solve(M3, hr_2)  # M4 = M3^-1 @ hr_2
        k_1 = sigma2_rho * k_0 + sigma2 * (
            TestMFK_variance.corr(x1, x2, sm, 2) - hr_1.T @ M4
        )
        return k_0, k_1

    @staticmethod
    def Cov_LG_LG(x1, x2, sm):
        # computation of the covariance  based on the formula from LeGratiet 2015
        # Using the Sigma2 output from Le Gratiet paper
        x_t_lf = sm.training_points[0][0][0]
        y_t_lf = sm.training_points[0][0][1]
        n_LF = x_t_lf.shape[0]
        (
            beta_LG_1,
            sigma2_LG_1,
            beta_LG_2,
            sigma2_LG_2,
            rho_LG,
            sigma2_rho_LG,
        ) = TestMFK_variance.hyperparam_LG(sm)
        sigma2 = sigma2_LG_1
        C = sm.optimal_par[0]["C"]
        R = C @ C.T
        r_x1 = np.empty((n_LF, 1))
        r_x2 = np.empty((n_LF, 1))
        for i in range(n_LF):
            r_x1[i, 0] = TestMFK_variance.corr(x1, x_t_lf[i], sm, 1)
            r_x2[i, 0] = TestMFK_variance.corr(x2, x_t_lf[i], sm, 1)
        h_x1 = np.ones((1, 1))
        h_x2 = np.ones((1, 1))
        H = np.ones((n_LF, 1))
        hr_1 = np.vstack((h_x1, r_x1))
        hr_2 = np.vstack((h_x2, r_x2))
        M1 = np.hstack((np.zeros((1, 1)), H.T))
        M2 = np.hstack((H, R))
        M3 = np.vstack((M1, M2))
        M4 = np.linalg.solve(M3, hr_2)  # M4 = M3^-1 @ hr_2
        k_0 = sigma2 * (
            TestMFK_variance.corr(x1, x2, sm, 1) - hr_1.T @ M4
        )  # covariance of level 0

        x_t_hf = sm.training_points[None][0][0]
        n_HF = x_t_hf.shape[0]
        sigma2 = sigma2_LG_2
        C = sm.optimal_par[1]["C"]
        R = C @ C.T
        r_x1 = np.empty((n_HF, 1))
        r_x2 = np.empty((n_HF, 1))
        for i in range(n_HF):
            r_x1[i, 0] = TestMFK_variance.corr(x1, x_t_hf[i], sm, 2)
            r_x2[i, 0] = TestMFK_variance.corr(x2, x_t_hf[i], sm, 2)
        mu_x1 = sm._predict_intermediate_values(x1, 1).reshape(-1, 1)
        mu_x2 = sm._predict_intermediate_values(x2, 1).reshape(-1, 1)
        h_x1 = np.vstack((mu_x1, np.ones((1, 1))))
        h_x2 = np.vstack((mu_x2, np.ones((1, 1))))
        y_D_l = y_t_lf[-n_HF:, :].reshape(
            -1, 1
        )  # to get y^l-1(D^l) we need n_HF last points of y_t_lf
        H = np.hstack((y_D_l, np.ones((n_HF, 1))))
        hr_1 = np.vstack((h_x1, r_x1))
        hr_2 = np.vstack((h_x2, r_x2))
        M1 = np.hstack((np.zeros((2, 2)), H.T))
        M2 = np.hstack((H, R))
        M3 = np.vstack((M1, M2))
        M4 = np.linalg.solve(M3, hr_2)  # M4 = M3^-1 @ hr_2

        k_1 = sigma2_rho_LG * k_0 + sigma2 * (
            TestMFK_variance.corr(x1, x2, sm, 2) - hr_1.T @ M4
        )
        return k_0, k_1

    @staticmethod
    def verif_hyperparam(sm, x_test_LHS):
        # get the hyperparameters from sm
        beta_sm_1 = sm.optimal_par[0]["beta"][0, 0]
        sigma2_sm_1 = sm.optimal_par[0]["sigma2"][0]
        rho_beta_sm = sm.optimal_par[1]["beta"]
        beta_sm_2 = rho_beta_sm[1, 0]
        sigma2_sm_2 = sm.optimal_par[1]["sigma2"][0]
        rho_sm = rho_beta_sm[0, 0]
        (var_all_pred, sigma2_rho) = sm.predict_variances_all_levels(x_test_LHS)
        sigma2_rho_sm = sigma2_rho[0]

        (
            beta_LG_1,
            sigma2_LG_1,
            beta_LG_2,
            sigma2_LG_2,
            rho_LG,
            sigma2_rho_LG,
        ) = TestMFK_variance.hyperparam_LG(sm)

        return (
            beta_sm_1,
            sigma2_sm_1,
            beta_sm_2,
            sigma2_sm_2,
            rho_sm,
            sigma2_rho_sm,
            beta_LG_1,
            sigma2_LG_1,
            beta_LG_2,
            sigma2_LG_2,
            rho_LG,
            sigma2_rho_LG,
        )

    def test_mfk_variance(self):
        # To create the doe
        # dim = 2
        nlevel = 2
        ub0 = 10.0
        ub1 = 15.0
        lb0 = -5.0
        lb1 = 0.0
        xlimits = np.array([[lb0, ub0], [lb1, ub1]])

        # Constants
        n_HF = 5  # number of high fidelity points (number of low fi is twice)
        xdoes = NestedLHS(nlevel=nlevel, xlimits=xlimits)
        x_t_lf, x_t_hf = xdoes(n_HF)

        # Evaluate the HF and LF functions
        y_t_lf = LF(x_t_lf)
        y_t_hf = HF(x_t_hf)

        sm = MFK(
            theta0=x_t_hf.shape[1] * [1e-2],
            print_global=False,
            rho_regr="constant",
        )

        # low-fidelity dataset names being integers from 0 to level-1
        sm.set_training_values(x_t_lf, y_t_lf, name=0)
        # high-fidelity dataset without name
        sm.set_training_values(x_t_hf, y_t_hf)
        # train the model
        sm.train()

        # Validation set
        # for validation with LHS
        ntest = 1
        sampling = LHS(xlimits=xlimits)
        x_test_LHS = sampling(ntest)
        # y_test_LHS = HF(x_test_LHS)

        # compare the mean value between different formula
        if print_output:
            print("Mu sm  : {}".format(sm.predict_values(x_test_LHS)[0, 0]))
            print(
                "Mu LG_sm : {}".format(TestMFK_variance.Mu_LG_sm(x_test_LHS, sm)[0, 0])
            )
            print(
                "Mu LG_LG : {}".format(TestMFK_variance.Mu_LG_LG(x_test_LHS, sm)[0, 0])
            )

        # self.assertAlmostEqual(
        #     TestMFK_variance.Mu_LG_sm(x_test_LHS, sm)[0, 0],
        #     TestMFK_variance.Mu_LG_LG(x_test_LHS, sm)[0, 0],
        #     delta=1,
        # )
        self.assertAlmostEqual(
            sm.predict_values(x_test_LHS)[0, 0],
            TestMFK_variance.Mu_LG_LG(x_test_LHS, sm)[0, 0],
            delta=1,
        )

        # compare the variance value between different formula
        (k_0_LG_sm, k_1_LG_sm) = TestMFK_variance.Cov_LG_sm(x_test_LHS, x_test_LHS, sm)
        (k_0_LG_LG, k_1_LG_LG) = TestMFK_variance.Cov_LG_LG(x_test_LHS, x_test_LHS, sm)
        k_0_sm = sm.predict_variances_all_levels(x_test_LHS)[0][0, 0]
        k_1_sm = sm.predict_variances_all_levels(x_test_LHS)[0][0, 1]

        if print_output:
            print("Level 0")
            print("Var sm  : {}".format(k_0_sm))
            print("Var LG_sm : {}".format(k_0_LG_sm[0, 0]))
            print("Var LG_LG : {}".format(k_0_LG_LG[0, 0]))

            print("Level 1")
            print("Var sm  : {}".format(k_1_sm))
            print("Var LG_sm : {}".format(k_1_LG_sm[0, 0]))
            print("Var LG_LG : {}".format(k_1_LG_LG[0, 0]))

        # for level 0
        self.assertAlmostEqual(k_0_sm, k_0_LG_sm[0, 0], delta=1)
        self.assertAlmostEqual(k_0_LG_sm[0, 0], k_0_LG_LG[0, 0], delta=1)
        # for level 1
        self.assertAlmostEqual(k_1_sm, k_1_LG_sm[0, 0], delta=1)
        self.assertAlmostEqual(k_1_LG_sm[0, 0], k_1_LG_LG[0, 0], delta=1)

        (
            beta_sm_1,
            sigma2_sm_1,
            beta_sm_2,
            sigma2_sm_2,
            rho_sm,
            sigma2_rho_sm,
            beta_LG_1,
            sigma2_LG_1,
            beta_LG_2,
            sigma2_LG_2,
            rho_LG,
            sigma2_rho_LG,
        ) = TestMFK_variance.verif_hyperparam(sm, x_test_LHS)
        if print_output:
            print("Hyperparameters")
            print("rho_sm : {}".format(rho_sm))
            print("rho_LG : {}".format(rho_LG))
            print("sigma2_rho_sm : {}".format(sigma2_rho_sm[0]))
            print("sigma2_rho_LG : {}".format(sigma2_rho_LG))
            print("beta_sm_1 : {}".format(beta_sm_1))
            print("beta_LG_1 : {}".format(beta_LG_1[0, 0]))
            print("beta_sm_2 : {}".format(beta_sm_2))
            print("beta_LG_2 : {}".format(beta_LG_2))
            print("sigma2_sm_1 : {}".format(sigma2_sm_1))
            print("sigma2_LG_1 : {}".format(sigma2_LG_1))
            print("sigma2_sm_2 : {}".format(sigma2_sm_2))
            print("sigma2_LG_2 : {}".format(sigma2_LG_2))


if __name__ == "__main__":
    unittest.main()



try:
    import matplotlib

    matplotlib.use("Agg")
    NO_MATPLOTLIB = False
except ImportError:
    NO_MATPLOTLIB = True

import unittest

import numpy as np

from smt.problems import TensorProduct
from smt.sampling_methods import LHS

# from smt.utils.misc import compute_relative_error
# from smt.utils.silence import Silence
from smt.utils.sm_test_case import SMTestCase
from smt.applications.mfck import MFCK

print_output = False


class TestMFKOneFidelity(SMTestCase):
    def setUp(self):
        self.nt = 20
        self.ne = 50
        self.ndim = 1

    def test_mfk_1fidelity(self):
        self.problems = ["gaussian"]

        for fname in self.problems:
            prob = TensorProduct(ndim=self.ndim, func=fname)
            sampling = LHS(xlimits=prob.xlimits, random_state=0)

            np.random.seed(0)
            xt = sampling(self.nt)
            yt = prob(xt)
            for i in range(self.ndim):
                yt = np.concatenate((yt, prob(xt, kx=i)), axis=1)

            sampling = LHS(xlimits=prob.xlimits, random_state=0)

            sm1 = MFCK(hyper_opt="Cobyla")

            sm1.set_training_values(xt, yt[:, 0])
            sm1.train()

            mean, cov = sm1._predict(xt)

            num = np.linalg.norm(mean[:, 0] - yt[:, 0])
            den = np.linalg.norm(yt[:, 0])

            t_error = num / den

            self.assert_error(t_error, 0.0, 1e-4, 1e-4)

    @staticmethod
    def run_mfck_example_1fidelity():
        import matplotlib.pyplot as plt
        import numpy as np

        from smt.applications.mfk import NestedLHS
        from smt.applications.mfck import MFCK

        # Consider only 1 fidelity level
        # high fidelity model
        def hf_function(x):
            import numpy as np

            return ((x * 6 - 2) ** 2) * np.sin((x * 6 - 2) * 2)

        # Problem set up
        xlimits = np.array([[0.0, 1.0]])
        xdoes = NestedLHS(nlevel=1, xlimits=xlimits, random_state=0)
        xt_e = xdoes(7)[0]

        # Evaluate the HF function
        yt_e = hf_function(xt_e)

        sm1 = MFCK(theta0=[1.0], hyper_opt="Cobyla")

        # High-fidelity dataset without name
        sm1.set_training_values(xt_e, yt_e)

        # Train the model
        sm1.train()

        x = np.linspace(0, 1, 101, endpoint=True).reshape(-1, 1)

        # Query the outputs
        y, cov = sm1._predict(x)
        # _mse = sm.predict_variances(x)
        # _derivs = sm.predict_derivatives(x, kx=0)

        plt.figure()

        plt.plot(x, hf_function(x), label="reference")
        plt.plot(x, y, linestyle="-.", label="mean_gp")
        plt.scatter(xt_e, yt_e, marker="o", color="k", label="HF doe")

        plt.legend(loc=0)
        plt.ylim(-10, 17)
        plt.xlim(-0.1, 1.1)
        plt.xlabel(r"$x$")
        plt.ylabel(r"$y$")

        plt.show()

    # run scripts are used in documentation as documentation is not always rebuild
    # make a test run by pytest to test the run scripts
    @unittest.skipIf(NO_MATPLOTLIB, "Matplotlib not installed")
    def test_run_mfck_example_1fidelity(self):
        self.run_mfck_example_1fidelity()


if __name__ == "__main__":
    unittest.main()



from smt.design_space.design_space import (
    CategoricalVariable,
    BaseDesignSpace,
    FloatVariable,
    IntegerVariable,
    OrdinalVariable,
    DesignSpace,
    DesignVariable,
    ensure_design_space,
)

__all__ = [
    "BaseDesignSpace",
    "FloatVariable",
    "IntegerVariable",
    "OrdinalVariable",
    "CategoricalVariable",
    "DesignSpace",
    "DesignVariable",
    "ensure_design_space",
]



from .kernels import (
    Kernel,
    PowExp,
    SquarSinExp,
    Matern32,
    Matern52,
    ActExp,
    Constant,
    Operator,
    Sum,
    Product,
)

__all__ = [
    "Kernel",
    "PowExp",
    "SquarSinExp",
    "Matern32",
    "Matern52",
    "ActExp",
    "Constant",
    "Operator",
    "Sum",
    "Product",
]
